from collections import defaultdict
import functools
import dis
import logging
from threading import Condition
import time
import multiprocessing
import sys
import os
import json
import threading
import pty
import collections
import subprocess
import signal
from typing import (
    DefaultDict,
    Optional,
    Tuple,
    Dict,
    Callable,
    Union,
    List,
    Any,
)
from logging.handlers import DatagramHandler
from websocket_server import WebsocketServer  # type: ignore
from py2cfg.builder import CFGBuilder
from py2cfg._source_db import CFG_Database


# --- Forward type declarations --- #
# Stores pid of debugger/this process
_pid: int

# Stores pid of fork
_child_pid: int

LineCache = DefaultDict[
    Tuple[str, str, int],  # filename, funcname, stacklevel
    List[int],  # list of executed lines
]

logger = logging.getLogger(__name__)

LOCAL_SL = "__py2cfg_stacklevel"

# Cheat sheet on stack frame objects.
# https://docs.python.org/3/library/inspect.html


# --- Server handlers --- #

_cond = threading.Condition()


def new_client(client, server):
    cfg = CFGBuilder().build_from_src("", "")
    _path = cfg.build_visual(
        "key",
        format="svg",
        show=False,
        calls=True,
        cleanup=True,
        directory=os.path.join(os.environ["PY2CFG_SERVE_AT"], "cfg"),
        build_keys=True,
    )
    logger.debug(f"Build key at {_path}")
    server.send_message(
        client,
        json.dumps(
            {"type": "key", "filepath": os.path.join("cfg", "key.svg"),}
        ),
    )

    with _cond:
        _cond.notify_all()


def client_left(client, server):
    # How to gracefully shutdown pudb when browser tab is closed?
    pass


def message_received(client, server, message):
    """Called when SVG link is clicked ?"""


server = WebsocketServer(int(os.environ["PY2CFG_SOCKET_PORT"]))
server.set_fn_new_client(new_client)
server.set_fn_client_left(client_left)
server.set_fn_message_received(message_received)

_entry_frame = None


def _set_entry_frame(frame, name):
    global _entry_frame
    if frame.f_code.co_filename == name and _entry_frame is None:
        _entry_frame = frame
        frame.f_locals[LOCAL_SL] = 0
        with _cond:
            # Wait so a browser can service the request
            _cond.wait_for(lambda: server.clients)
        return 0
    return -1


def _get_stack_level(current_frame):
    assert _entry_frame is not None
    last_visited_frame = None
    frame = current_frame
    delta_sl = 0
    while frame:
        if LOCAL_SL in frame.f_locals:
            last_visited_frame = frame
            break

        frame = frame.f_back
        delta_sl += 1

    frame_sl = delta_sl + last_visited_frame.f_locals[LOCAL_SL]
    frame = current_frame
    while frame and frame_sl:
        if LOCAL_SL not in frame.f_locals:
            frame.f_locals[LOCAL_SL] = frame_sl
        else:
            break
        frame_sl -= 1

    return current_frame.f_locals[LOCAL_SL]


# --- Databases --- #
cfg_db = CFG_Database()


def wrap_user_call(func, entry):
    def user_call(dbg, _frame, argument_list):
        if _pid == os.getpid():
            _set_entry_frame(_frame, entry)

        return func(dbg, _frame, argument_list)

    return user_call


_child_pid = None  # type: ignore


def get_future_id():
    return _child_pid


class ExecData(tuple):
    @property
    def filename(self):
        return self[0]

    @property
    def funcname(self):
        return self[1]

    @property
    def lineno(self):
        return self[2]

    @property
    def event(self):
        return self[3]

    @property
    def stack_level(self):
        return self[4]

    @classmethod
    def create(
        cls,
        filename: str,
        funcname: str,
        lineno: int,
        event: str,
        stacklevel: int,
    ):
        return cls((filename, funcname, lineno, event, stacklevel))

    @classmethod
    def from_frame(cls, frame, event):
        return cls(
            (
                frame.f_code.co_filename,
                frame.f_code.co_name,
                frame.f_lineno,
                event,
                _get_stack_level(frame),
            )
        )


def trace_create(f_id, queue_object):
    def trace(frame, event, args):
        queue_object.put(ExecData.from_frame(frame, event))
        return trace

    return trace


class FrameWrapper:
    def __init__(self, frame, rel=_entry_frame):
        assert _entry_frame is not None
        self._frame = frame
        _get_stack_level(frame)

    @property
    def name(self):
        return self._frame.f_code.co_name

    @property
    def frame(self):
        return self._frame

    @property
    def stack_level(self):
        return self._frame.f_locals[LOCAL_SL]

    @property
    def filename(self):
        return self._frame.f_code.co_filename

    @property
    def lineno(self):
        return self._frame.f_lineno

    def settrace(self, tracefn, stop=None):
        sys.settrace(tracefn)
        frame = self._frame
        while frame is not None:
            frame.f_trace = tracefn
            if frame == stop:
                break
            frame = frame.f_back
        return frame

    _calls_at_line: Dict[
        Tuple[str, int], Optional[bool],
    ] = collections.defaultdict(lambda: None)

    def calls_function(self):
        if (res := self._calls_at_line[self]) is not None:
            return res
        generator = iter(dis.Bytecode(self.frame.f_code))
        instruction = next(generator)
        try:
            while True:
                instruction = next(generator)
                if instruction.starts_line == self.frame.f_lineno:
                    if "CALL" in instruction.opname:
                        self._calls_at_line[self] = True
                        return True
                    while (instruction := next(generator)).starts_line is None:
                        if "CALL" in instruction.opname:
                            self._calls_at_line[self] = True
                            return True
        except StopIteration:
            self._calls_at_line[self] = False
            return False

    def __str__(self):
        frame = self._frame
        string = super().__str__() + "\n"
        while frame is not None:
            string += f"\t{frame.f_code.co_filename}:{frame.f_lineno}\n"
            frame = frame.f_back
        return string

    def __hash__(self):
        return hash((self.frame.f_code.co_filename, self.frame.f_lineno))


class QueueMGR:
    def __init__(self, mp_queue, factory, line_cache: LineCache) -> None:
        self._lock = threading.Lock()
        self._line_cache = line_cache
        self._rx = mp_queue
        self._running = True
        self.thread = threading.Thread(target=self._start_fn, daemon=True)
        self.thread.start()
        self._thread_ident = -1
        self._f_id = factory

        self._requests: Dict[
            Tuple[str, str, int], Union[Tuple[Condition, List[int]], list]
        ] = {}

    def set_thread_ident(self, ident):
        with self._lock:
            self._thread_ident = ident

    def clear_thread_ident(self):
        with self._lock:
            self._thread_ident = -1

    def post(self, key: Tuple[str, str, int]):
        self._requests[key] = [threading.Condition(), None]

    def wait(self, key, timeout=None) -> list:
        assert key in self._requests
        cond = self._requests[key][0]
        with cond:
            cond.wait_for(lambda: self._requests[key][1] is not None, timeout)
            lines = self._requests[key][1]
            del self._requests[key]
        return lines

    def _start_fn(self):
        while self._running:
            try:
                data: ExecData = self._rx.get(timeout=0.1)
            except:
                time.sleep(0.01)
                continue
            else:
                key = data.filename, data.funcname, data.stack_level
                if key in self._requests:
                    cond = self._requests[key][0]
                    with cond:
                        self._requests[key][1] = self._line_cache[key]
                        cond.notify_all()

                if data.event == "call":
                    logger.debug(f"call: {data.funcname}")
                elif data.event == "return":
                    logger.debug(f"return: {data.funcname}")
                    self._line_cache[key] = []

                self._line_cache[key].append(data.lineno)

    def close(self) -> None:
        self._rx.close()
        self._running = False


class ProcessFactory:
    def __new__(
        cls,
        frame: FrameWrapper,
        func: Callable,
        stop: Any,
        continue_fn: Callable = None,
    ):
        ipc = multiprocessing.Queue()  # type: ignore
        cond = multiprocessing.Condition()

        # First fork saves the context of 'frame' before we enter a function
        # at the debugger.
        # TODO Don't fork if there isn't a function call at current line.
        if (factory := os.fork()) == 0:
            pid = os.getpid()
            logger.info(f"Factory {pid} created at level: {frame.stack_level}")
            last_child = -1
            while True:
                # Wait for parent to signal this process.
                os.kill(pid, signal.SIGSTOP)
                time.sleep(0.01)
                child, _ = pty.fork()  # Don't inherit stdin/stdout descriptors
                if child == 0:
                    # Second fork starts execution from saved context. A child
                    # is created on every step. The last child created is killed
                    time.sleep(0.01)
                    # Install trace handlers
                    if continue_fn is not None:
                        continue_fn()

                    trace = trace_create(pid, ipc)
                    return func(frame.settrace(trace, stop))
                else:
                    if last_child != -1:
                        os.kill(last_child, signal.SIGTERM)
                        _, status = os.waitpid(last_child, 0)
                    else:
                        status = -1
                    logger.info(
                        f"\n\tFactory ({pid}) stack level {frame.stack_level}: "
                        f"{frame.filename}:{frame.lineno}"
                        f"\n\t-> Fork: {child}"
                        f"\n\t Kill: {last_child} status: {status}\n"
                    )
                    last_child = child
                    continue
        else:
            cache = defaultdict(list)  # type: ignore
            obj = super().__new__(cls)
            # __init__()
            obj._frame = frame
            obj._pid = factory
            obj._queue = QueueMGR(ipc, factory, cache)
            obj._stack_level = frame.stack_level
            obj._line_cache = cache
            return obj

    def draw_graph(self, frame: FrameWrapper, create_child=True) -> None:
        key = frame.filename, frame.name, frame.stack_level
        if self.line_cache[key]:
            return threading.Thread(
                target=self._draw_graph,
                args=(frame.frame, self.line_cache[key]),
                daemon=True,
            ).start()

        if create_child:
            threading.Thread(
                target=self._draw_graph,
                args=(frame.frame, self.line_cache[key]),
                daemon=True,
            ).start()

            self.queue.post(key)
            os.kill(self.pid, signal.SIGCONT)

            def _start_fn():
                time.sleep(1)
                self.line_cache[key] = lines = self.queue.wait(key)
                last_size = 0
                while (size := len(lines)) != last_size:
                    if threading.current_thread().ident != self.thrd_ident:
                        logger.warn("Thread ident changed. Cancelling build")
                        break
                    last_size = size
                    self._draw_graph(frame.frame, lines)
                    time.sleep(1)

            thrd = threading.Thread(target=_start_fn, daemon=True,)
            thrd.start()
            self.thrd_ident = thrd.ident

    def terminate(self) -> int:
        os.kill(self.pid, signal.SIGTERM)
        os.kill(self.pid, signal.SIGCONT)
        _, status = os.waitpid(self.pid, 0)
        self.queue.close()
        logger.info(
            f"\n\tKill {self.pid}"
            f"\n\tlevel {self.stack_level}"
            f"\n\tstatus {status}"
            f"\n"
        )
        return status

    @property
    def frame(self) -> FrameWrapper:
        return self._frame  # type: ignore

    @property
    def stack_level(self) -> int:
        return self._stack_level  # type: ignore

    @property
    def pid(self) -> int:
        return self._pid  # type: ignore

    @property
    def queue(self) -> QueueMGR:
        return self._queue  # type: ignore

    @property
    def line_cache(self) -> LineCache:
        return self._line_cache  # type: ignore

    nbuild = 0

    def _draw_graph(self, frame, lst):
        name = frame.f_globals["__name__"]
        if frame.f_code.co_name != "<module>":
            name = f"{name}.{frame.f_code.co_name}"

        key = frame.f_code.co_filename, frame.f_code.co_name
        cfg, context_blocks = cfg_db.at_function[key]

        if cfg.outline_block(frame.f_lineno, context_blocks) < 0:
            logger.error(f"Unable to find block at line {frame.f_lineno}")

        cfg.highlight_blocks(lst, context_blocks)
        dispatch_cfg(cfg, f"{name}{self.nbuild}", True)
        self.nbuild += 1


class ProcessStack(collections.deque):
    curr_stack_level: int = 0
    base_factory: ProcessFactory = None  # type: ignore


_process_stack = ProcessStack()


def dispatch_cfg(cfg, name=None, interactive=False):
    name = cfg.name if name is None else name
    try:
        src = cfg.build_visual(
            name,
            format="svg",
            show=False,
            calls=True,
            cleanup=True,
            directory=os.path.join(os.environ["PY2CFG_SERVE_AT"], "cfg"),
            interactive=interactive,
            build_keys=False,
        )
    except:
        logger.warn("Failed to build cfg from source")
    else:
        server.send_message_to_all(
            json.dumps(
                {
                    "type": "cfg",
                    "filepath": os.path.join("cfg", f"{name}.svg"),
                    "layer": 0,
                }
            )
        )
        logger.debug(f"Created at {src}")


def wrap_user_line(func, entry):
    name = os.path.basename(os.path.splitext(entry)[0])

    def user_line(dbg, _frame):
        # Only fork if this is the main debugger process...
        if _pid == os.getpid() and _entry_frame is not None:
            frame = FrameWrapper(_frame)
            if frame.stack_level >= _process_stack.curr_stack_level:
                if (
                    factory := ProcessFactory(
                        frame, functools.partial(func, dbg), dbg.botframe,
                    )
                ) is None:
                    # Continue program execution from caller's stack frame.
                    return

                if frame.stack_level > _process_stack.curr_stack_level:
                    # Entered a function call on step. The factory is forked
                    # if next step enters a higher stack level
                    _process_stack.appendleft(factory)

                elif (
                    _process_stack.curr_stack_level == frame.stack_level
                    and _process_stack[0] is None
                ):
                    _process_stack[0] = factory
                    _process_stack.base_factory = ProcessFactory(
                        frame, functools.partial(func, dbg), dbg.botframe,
                    )
                    if _process_stack.base_factory is None:
                        # Continue program execution from first line of program
                        return

                    cfg = CFGBuilder(short=True).build_from_file(name, entry)
                    cfg_db.index(entry, cfg)

                    try:
                        src = cfg.build_visual(
                            name,
                            format="svg",
                            show=False,
                            calls=True,
                            cleanup=True,
                            directory=os.path.join(
                                os.environ["PY2CFG_SERVE_AT"], "cfg"
                            ),
                            build_keys=False,
                        )
                    except:
                        logger.warn("Failed to build cfg from source")
                    else:
                        server.send_message_to_all(
                            json.dumps(
                                {
                                    "type": "cfg",
                                    "filepath": os.path.join(
                                        "cfg", f"{name}.svg"
                                    ),
                                    "layer": 0,
                                }
                            )
                        )
                        logger.debug(f"Created at {src}")
                    return func(dbg, _frame)

                elif _process_stack.curr_stack_level == frame.stack_level:
                    assert _process_stack[0].pid > 0
                    _process_stack[0].terminate()
                    _process_stack[0] = factory

            elif frame.stack_level < _process_stack.curr_stack_level:
                # Since we can jump back any number of stack frames, clean up
                # all factories at stack levels higher than this one.
                while True:
                    if _process_stack[0].stack_level <= frame.stack_level:
                        break
                    assert _process_stack[0].pid > 0
                    _process_stack[0].terminate()
                    del _process_stack[0]

            if (k := (frame.filename, frame.lineno)) in cfg_db.at_lineno:
                # When stopped at a functionDef statement, generate a
                # cfg with normal styling
                subcfg, _ = cfg_db.at_lineno[k]
                logger.warn(f"{str(subcfg)} @ {frame.filename}:{frame.lineno}")
                threading.Thread(
                    target=dispatch_cfg, args=(subcfg,), daemon=True,
                ).start()

            elif (size := len(_process_stack)) > 1:
                # Fork previous stack frame's factory
                for i in range(1, size):
                    caller = _process_stack[i]
                    if caller.stack_level < frame.stack_level:
                        caller.draw_graph(
                            frame,
                            frame.stack_level > _process_stack.curr_stack_level,
                        )
                        break

            elif _process_stack.curr_stack_level == 0:
                assert _process_stack.base_factory is not None
                _process_stack.base_factory.draw_graph(frame, False)

            _process_stack.curr_stack_level = frame.stack_level
        return func(dbg, _frame)

    return user_line


def wrap_user_return(func, entry):
    def user_return(dbg, frame, return_value):
        if _pid == os.getpid():
            _set_entry_frame(frame, entry)
            if _entry_frame is not None:
                subcfg, blocks = cfg_db.at_function[
                    frame.f_code.co_filename, frame.f_code.co_name
                ]
                if subcfg.outline_block(frame.f_lineno, blocks) < 0:
                    logger.warn(
                        f"Failed to highlight block at {frame.f_lineno}"
                    )
                threading.Thread(
                    target=dispatch_cfg,
                    args=(subcfg, subcfg.name, True),
                    daemon=True,
                ).start()

            logger.debug(f"return: {frame.f_code.co_name}")
        return func(dbg, frame, return_value)

    return user_return


def wrap_user_exception(func, entry):
    def user_exception(dbg, frame, exc_tuple):
        if _pid == os.getpid():
            logger.info(
                f"exception: {frame.f_code.co_filename}:{frame.f_lineno}"
            )
        return func(dbg, frame, exc_tuple)

    return user_exception


class Handler(logging.Handler):
    def __init__(self, server):
        super().__init__()
        self.server = server

    def emit(self, record):
        self.server.send_message_to_all(
            json.dumps(
                {
                    "type": "log",
                    "data": self.format(record),
                    "levelname": record.levelname,
                }
            )
        )


def debug_init(debug_class, entryfile, port=None) -> None:
    global _pid
    port = 8000 if port is None else port
    _pid = os.getpid()
    _process_stack.appendleft(None)

    rootlog = logging.getLogger()

    handler = Handler(server)
    rootlog.addHandler(handler)
    handler.setFormatter(logging.Formatter("%(asctime)s - %(message)s"))
    logger.error("\033[H\033[J")
    logger.setLevel(getattr(logging, os.environ["PY2CFG_LOGLEVEL"], 0))

    # TOP SNEAKY: Swap whatever debugger functions that [DEBUG PROG] uses with
    # our own user_ functions.
    debug_class.user_call = wrap_user_call(debug_class.user_call, entryfile)
    debug_class.user_line = wrap_user_line(debug_class.user_line, entryfile)
    debug_class.user_return = wrap_user_return(
        debug_class.user_return, entryfile
    )
    debug_class.user_exception = wrap_user_exception(
        debug_class.user_exception, entryfile
    )
    threading.Thread(target=server.run_forever, daemon=True,).start()

    try:
        subprocess.Popen(["xdg-open", f"http://localhost:{port}"])
    except OSError:
        print(
            f"Could not start browser. Program will hang unless browser attaches to localhost:{port}"
        )
        return


__all__ = ["debug_init", "module_context", "get_future_id"]



import pickle
from .filewatcher import create_managed_database
from .db import query_filename_lineno, connect
from .util import getcwd
from flask import Flask, abort, request
import logging
import os
import base64

logger = logging.getLogger(__name__)
server = Flask(__name__)
create_managed_database()

REMOTE_CALLS = {}


def remote(func):
    REMOTE_CALLS[func.__name__] = func
    return func


def validate_data(data):
    def resolve(result):
        return {
            "jsonrpc": data["jsonrpc"],
            "id": data["id"],
            "result": result,
        }

    def failure(reason):
        return {
            "jsonrpc": data["jsonrpc"],
            "id": data["id"],
            "error": {"code": 1, "message": reason},
        }

    if data is None:
        return None, None
    if not all(s in data for s in ("jsonrpc", "method", "params", "id")):
        return None, None
    return resolve, failure


def strip_svg_doctype(svgdata: str) -> str:
    lines = svgdata.splitlines(keepends=True)

    def iter_lines():
        found = False
        for line in lines:
            if found:
                yield line
            else:
                if line.startswith("<svg"):
                    found = True
                    yield line

    return "".join(iter_lines())


@server.route("/")
def index():
    return "py2cfg"


@server.route("/build")
def build(method=["GET"]):
    data = request.get_json()
    resolve, failure = validate_data(data)
    if resolve is None:
        abort(400)

    logger.info(f"method: {data['method']}, params: {data['params']}")
    method = data["method"]
    params = data["params"]

    if method not in REMOTE_CALLS:
        abort(400)

    func = REMOTE_CALLS[method]
    return resolve(func(**params))


@remote
def build_graph(
    filename, lineno, build_own=True, build_keys=False, interactive=False
):
    with connect() as cursor:
        cfg, own_blocks = query_filename_lineno(filename, lineno, cursor)
        if interactive:
            cfg.outline_block(lineno, own_blocks)
            cfg.highlight_blocks([lineno], own_blocks)

        path = cfg.build_visual(
            "_tmp_file",
            "svg",
            show=False,
            directory=os.path.join(getcwd(), ".py2cfg"),
            build_own=build_own,
            build_keys=build_keys,
            interactive=interactive,
        )
        with open(path, "r") as fp:
            svg_data = fp.read()

        os.remove(path)
        return strip_svg_doctype(svg_data)


@remote
def build_cfg(filename, lineno):
    with connect() as cursor:
        cfg, _ = query_filename_lineno(filename, lineno, cursor)
        return str(base64.b64encode(pickle.dumps(cfg)), encoding="utf-8")



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
Control flow graph builder.
"""

from collections import defaultdict, deque
from typing import Dict, List, Optional, DefaultDict, Deque, Set, Union
import ast  # type: ignore
import enum
from _ast import (
    Import,
    Break,
    AnnAssign,
    For,
    ImportFrom,
    Assign,
    AugAssign,
    Call,
    ClassDef,
    Compare,
    Expr,
    FunctionDef,
    If,
    Module,
    Return,
    stmt,
    While,
    Yield,
    Try,
    ExceptHandler,
)  # type: ignore
from py2cfg.model import Block, TryBlock, Link, CFG, FuncBlock


def invert(
    node: Union[Compare, ast.expr]
) -> Union[ast.NameConstant, ast.UnaryOp, Compare]:
    """
    Invert the operation in an ast node object (get its negation).

    Args:
        node: An ast node object.

    Returns:
        An ast node object containing the inverse (negation) of the input node.
    """
    inverse = {
        ast.Eq: ast.NotEq,
        ast.NotEq: ast.Eq,
        ast.Lt: ast.GtE,
        ast.LtE: ast.Gt,
        ast.Gt: ast.LtE,
        ast.GtE: ast.Lt,
        ast.Is: ast.IsNot,
        ast.IsNot: ast.Is,
        ast.In: ast.NotIn,
        ast.NotIn: ast.In,
    }

    if isinstance(node, ast.Compare):
        op = type(node.ops[0])
        inverse_node: Union[
            ast.NameConstant, ast.UnaryOp, Compare
        ] = ast.Compare(
            left=node.left, ops=[inverse[op]()], comparators=node.comparators
        )
    elif isinstance(node, ast.NameConstant) and node.value in [True, False]:
        inverse_node = ast.NameConstant(value=not node.value)
    else:
        inverse_node = ast.UnaryOp(op=ast.Not(), operand=node)

    return inverse_node


def merge_exitcases(
    exit1: Union[Compare, ast.BoolOp, ast.expr, None],
    exit2: Union[Compare, ast.BoolOp, ast.expr, None],
) -> Union[Compare, ast.BoolOp, ast.expr, None]:
    """
    Merge the exitcases of two Links.

    Args:
        exit1: The exitcase of a Link object.
        exit2: Another exitcase to merge with exit1.

    Returns:
        The merged exitcases.
    """
    if exit1:
        if exit2:
            return ast.BoolOp(ast.And(), values=[exit1, exit2])
        return exit1
    return exit2


class TryEnum(enum.IntEnum):
    BODY = enum.auto()
    EXCEPT = enum.auto()
    ELSE = enum.auto()
    FINAL = enum.auto()


class TryStackObject:
    def __init__(
        self,
        try_block: TryBlock,
        after_block: Block,
        has_final: bool,
        iter_state: Optional[TryEnum] = None,
    ) -> None:
        self.try_block = try_block
        self.after_block = after_block
        self.has_final = has_final
        self.iter_state = iter_state

    @property
    def node(self) -> ast.Try:
        return self.try_block.statements[0]  # type: ignore


class CFGBuilder(ast.NodeVisitor):
    """
    Control flow graph builder.

    A control flow graph builder is an ast.NodeVisitor that can walk through
    a program's AST and iteratively build the corresponding CFG.
    """

    def __init__(
        self, short: bool = True, treebuf: DefaultDict[str, Deque] = None
    ) -> None:
        self.isShort = short
        self._callbuf: List[FuncBlock] = []
        self._treebuf = defaultdict(deque) if treebuf is None else treebuf

    @property
    def loop_stack(self):
        return self._treebuf["loop_stack"]

    @property
    def try_stack(self) -> Deque[TryStackObject]:
        return self._treebuf["try_stack"]

    # ---------- CFG building methods ---------- #
    def build(
        self, name: str, tree: Module, asynchr: bool = False, entry_id: int = 0
    ) -> CFG:
        """
        Build a CFG from an AST.

        Args:
            name: The name of the CFG being built.
            tree: The root of the AST from which the CFG must be built.
            async: Boolean indicating whether the CFG being built represents an
                   asynchronous function or not. When the CFG of a Python
                   program is being built, it is considered like a synchronous
                   'main' function.
            entry_id: Value for the id of the entry block of the CFG.

        Returns:
            The CFG produced from the AST.
        """
        self.cfg = CFG(name, asynchr=asynchr, short=self.isShort)
        # Tracking of the current block while building the CFG.
        self.current_id = entry_id
        self.current_block = self.new_block()
        self.cfg.entryblock = self.current_block
        # Actual building of the CFG is done here.
        self.visit(tree)
        self.clean_cfg(self.cfg.entryblock, set())
        return self.cfg

    def build_from_src(self, name: str, src: str) -> CFG:
        """
        Build a CFG from some Python source code.

        Args:
            name: The name of the CFG being built.
            src: A string containing the source code to build the CFG from.

        Returns:
            The CFG produced from the source code.
        """
        tree = ast.parse(src, mode="exec")
        cfg = self.build(name, tree)
        cfg.lineno = 1
        cfg.end_lineno = len(src.splitlines())
        return cfg

    def build_from_file(self, name: str, filepath: str) -> CFG:
        """
        Build a CFG from some Python source file.

        Args:
            name: The name of the CFG being built.
            filepath: The path to the file containing the Python source code
                      to build the CFG from.

        Returns:
            The CFG produced from the source file.
        """
        with open(filepath, "r") as src_file:
            src = src_file.read()
            return self.build_from_src(name, src)

    # ---------- Graph management methods ---------- #
    def new_block(self, statement=None) -> Block:
        """
        Create a new block with a new id.

        Returns:
            A Block object with a new unique id.
        """
        self.current_id += 1
        block = Block(self.current_id)
        if statement is not None:
            block.add_statement(statement)
        return block

    def new_func_block(self) -> FuncBlock:
        """
        Create a new function block with a new id.

        Returns:
            A FuncBlock object with a new unique id.
        """
        self.current_id += 1
        return FuncBlock(self.current_id)

    def new_try_block(self, statement=None):
        self.current_id += 1
        block = TryBlock(self.current_id)
        if statement is not None:
            block.add_statement(statement)
        return block

    def add_statement(self, block: Block, statement: Union[stmt, Call]) -> None:
        """
        Add a statement to a block.

        Args:
            block: A Block object to which a statement must be added.
            statement: An AST node representing the statement that must be
                       added to the current block.
        """
        block.statements.append(statement)

    def add_exit(
        self,
        block: Block,
        nextblock: Block,
        exitcase: Union[Compare, None, ast.BoolOp, ast.expr] = None,
    ) -> None:
        """
        Add a new exit to a block.

        Args:
            block: A block to which an exit must be added.
            nextblock: The block to which control jumps from the new exit.
            exitcase: An AST node representing the 'case' (or condition)
                      leading to the exit from the block in the program.
        """
        newlink = Link(block, nextblock, exitcase)  # type: ignore
        block.exits.append(newlink)
        nextblock.predecessors.append(newlink)

    def new_loopguard(self) -> Block:
        """
        Create a new block for a loop's guard if the current block is not
        empty. Links the current block to the new loop guard.

        Returns:
            The block to be used as new loop guard.
        """
        if self.current_block.is_empty() and len(self.current_block.exits) == 0:
            # If the current block is empty and has no exits, it is used as
            # entry block (condition test) for the loop.
            loopguard = self.current_block
        else:
            # Jump to a new block for the loop's guard if the current block
            # isn't empty or has exits.
            loopguard = self.new_block()
            self.add_exit(self.current_block, loopguard)

        return loopguard

    def new_classCFG(self, node: ClassDef, asynchr: bool = False) -> None:
        """
        Create a new sub-CFG for a class definition and add it to the
        class CFGs of the CFG being built.

        Args:
            node: The AST node containing the class definition.
            async: Boolean indicating whether the class for which the CFG is
                   being built is asynchronous or not.
        """
        self.current_id += 1
        # A new sub-CFG is created for the body of the class definition and
        # added to the class CFGs of the current CFG.
        class_body = ast.Module(body=node.body)
        class_builder = CFGBuilder(self.isShort, self._treebuf)
        self.cfg.classcfgs[node.name] = classcfg = class_builder.build(
            node.name, class_body, asynchr, self.current_id
        )
        classcfg.lineno = node.lineno
        classcfg.end_lineno = node.end_lineno  # type: ignore
        self.current_id = class_builder.current_id + 1

    def new_functionCFG(self, node: FunctionDef, asynchr: bool = False) -> None:
        """
        Create a new sub-CFG for a function definition and add it to the
        function CFGs of the CFG being built.

        Args:
            node: The AST node containing the function definition.
            async: Boolean indicating whether the function for which the CFG is
                   being built is asynchronous or not.
        """
        self.current_id += 1
        # A new sub-CFG is created for the body of the function definition and
        # added to the function CFGs of the current CFG.
        func_body = ast.Module(body=node.body)
        func_builder = CFGBuilder(self.isShort, self._treebuf)
        cfg = self.cfg.functioncfgs[node.name] = func_builder.build(
            node.name, func_body, asynchr, self.current_id
        )
        self.current_id = func_builder.current_id + 1
        cfg.lineno = node.lineno
        cfg.end_lineno = node.end_lineno  # type: ignore

    def clean_cfg(self, block: Block, visited: Set[Block]) -> None:
        """
        Remove the useless (empty) blocks from a CFG.

        Args:
            block: The block from which to start traversing the CFG to clean
                   it.
            visited: A list of blocks that already have been visited by
                     clean_cfg (recursive function).
        """
        # Don't visit blocks twice.
        if block in visited:
            return
        visited.add(block)

        # Empty blocks are removed from the CFG.
        if block.is_empty():
            for pred in block.predecessors:
                for exit in block.exits:
                    self.add_exit(
                        pred.source,
                        exit.target,
                        merge_exitcases(pred.exitcase, exit.exitcase),
                    )
                    # Check if the exit hasn't yet been removed from
                    # the predecessors of the target block.
                    if exit in exit.target.predecessors:
                        exit.target.predecessors.remove(exit)
                # Check if the predecessor hasn't yet been removed from
                # the exits of the source block.
                if pred in pred.source.exits:
                    pred.source.exits.remove(pred)

            block.predecessors = []
            for exit in block.exits:
                self.clean_cfg(exit.target, visited)
            block.exits = []
        else:
            for exit in block.exits:
                self.clean_cfg(exit.target, visited)

    # ---------- AST Node visitor methods ---------- #
    def visit_ClassDef(self, node: ClassDef) -> None:
        self.add_statement(self.current_block, node)
        self.new_classCFG(node, asynchr=False)

    def visit_Expr(self, node: Expr) -> None:
        self.add_statement(self.current_block, node)
        self.generic_visit(node)

    def visit_Call(self, node: Call) -> None:
        def visit_func(node):
            if isinstance(node, ast.Name):
                return node.id
            elif isinstance(node, ast.Attribute):
                # Recursion on series of calls to attributes.
                func_name = visit_func(node.value)
                func_name += "." + node.attr
                return func_name
            elif isinstance(node, ast.Subscript):
                if "id" in node.value._fields:
                    return node.value.id
                elif "attr" in node.value._fields:
                    return visit_func(node.value)
                else:
                    raise AttributeError(
                        "WTF is this thing, build it in??", type(node)
                    )
            elif isinstance(node, ast.Constant):
                return node.value
            elif isinstance(node, ast.Call):
                if "id" in node.func._fields:
                    return node.func.id
                elif "attr" in node.func._fields:
                    return visit_func(node.func)
                else:
                    raise AttributeError(
                        "WTF is this thing, build it in??", type(node)
                    )
            else:
                print("WTF is this thing, build it in??", type(node))

        func = node.func
        func_name = visit_func(func)
        if isinstance(node, ast.Call):
            func_block = self.new_func_block()
            self.add_statement(func_block, node)
            func_block.name = func_name

            if self._callbuf:
                # Func block is argument of last block in self._callbuf
                self._callbuf[-1].args.append(func_block)
            else:
                # Not inside argument context.
                self.current_block.func_calls.append(func_name)
                self.current_block.func_blocks.append(func_block)

            self._callbuf.append(func_block)
            for arg in node.args:
                self.visit(arg)
            top = self._callbuf.pop()
            assert hash(top) == hash(func_block)
        else:
            self.current_block.func_calls.append(func_name)

    def visit_Assign(self, node: Assign) -> None:
        self.add_statement(self.current_block, node)
        self.generic_visit(node)

    def visit_AnnAssign(self, node: AnnAssign) -> None:
        self.add_statement(self.current_block, node)
        self.generic_visit(node)

    def visit_AugAssign(self, node: AugAssign) -> None:
        self.add_statement(self.current_block, node)
        self.generic_visit(node)

    def visit_Raise(self, node):
        if self.current_block.statements:
            raise_block = self.new_block(node)
            self.current_block.add_exit(raise_block)
            self.current_block = raise_block
        else:
            self.current_block.add_statement(node)

        if not self.try_stack:
            # Raise statement outside of try block
            # If we don't know where control jumps, this is the last block
            self.current_block = self.new_block(node)
            return

        if isinstance(node.exc, ast.Call):
            e_id = node.exc.func.id
        elif isinstance(node.exc, ast.Name):
            e_id = node.exc.id
        else:
            raise ValueError(f"Unexpected object {node.exc}")

        for tryobj in list(self.try_stack):

            def contains(item, state):
                return (
                    item in tryobj.try_block.except_blocks
                    and tryobj.iter_state == state
                )

            if contains(e_id, TryEnum.BODY):
                # try:
                #     raise StopIteration
                # except StopIteration:
                #     control_transfer_here = True
                # except Exception:
                #     control_transfer_here = False
                self.current_block.add_exit(
                    tryobj.try_block.except_blocks[e_id]
                )
                break
            elif contains(None, TryEnum.BODY):
                # try:
                #     raise StopIteration
                # except:
                #     control_transfers_here = 1
                self.current_block.add_exit(
                    tryobj.try_block.except_blocks[None]
                )
                break

            elif contains(e_id, TryEnum.EXCEPT) or contains(
                None, TryEnum.EXCEPT
            ):
                if tryobj.has_final:
                    _after_block = self.new_block()
                    self.current_block.add_exit(_after_block)
                    _after_block.add_exit(self.current_block)
                    self.current_block = _after_block

                    for child in tryobj.node.finalbody:
                        if isinstance(child, ast.Raise):
                            break
                        self.visit(child)

            elif tryobj.iter_state == TryEnum.ELSE:
                if tryobj.has_final:
                    _after_block = self.new_block()
                    self.current_block.add_exit(_after_block)
                    _after_block.add_exit(self.current_block)
                    self.current_block = _after_block
                    for child in tryobj.node.finalbody:
                        if isinstance(child, ast.Raise):
                            break
                        self.visit(child)

            elif tryobj.iter_state != TryEnum.FINAL and tryobj.has_final:
                # try: raise Exception
                # finally: pass
                _after_block = self.new_block()
                self.current_block.add_exit(_after_block)
                self.current_block = _after_block

                for child in tryobj.node.finalbody:
                    if isinstance(child, ast.Raise):
                        top = self.try_stack.popleft()
                        break
                    self.visit(child)
        self.current_block = self.new_block()

    def visit_Assert(self, node):
        self.add_statement(self.current_block, node)
        # New block for the case in which the assertion 'fails'.
        failblock = self.new_block()
        self.add_exit(self.current_block, failblock, invert(node.test))
        # If the assertion fails, the current flow ends, so the fail block is a
        # final block of the CFG.
        self.cfg.finalblocks.append(failblock)
        # If the assertion is True, continue the flow of the program.
        successblock = self.new_block()
        self.add_exit(self.current_block, successblock, node.test)
        self.current_block = successblock
        self.generic_visit(node)

    def visit_If(self, node: If) -> None:
        # If it already has something in it, we make a new block
        if self.current_block.statements:
            # Add the If statement at the beginning of the new block.
            cond_block = self.new_block()
            self.add_statement(cond_block, node)
            self.add_exit(self.current_block, cond_block)
            self.current_block = cond_block
        else:
            # Add the If statement at the end of the current block.
            self.add_statement(self.current_block, node)
        if any(isinstance(node.test, T) for T in (ast.Compare, ast.Call)):
            self.visit(node.test)
        # Create a new block for the body of the if. (storing the True case)
        if_block = self.new_block()

        self.add_exit(self.current_block, if_block, node.test)

        # Create a block for the code after the if-else.
        afterif_block = self.new_block()

        # New block for the body of the else if there is an else clause.
        if node.orelse:
            else_block = self.new_block()
            self.add_exit(self.current_block, else_block, invert(node.test))
            self.current_block = else_block
            # Visit the children in the body of the else to populate the block.
            for child in node.orelse:
                self.visit(child)
            self.add_exit(self.current_block, afterif_block)
        else:
            self.add_exit(self.current_block, afterif_block, invert(node.test))

        # Visit children to populate the if block.
        self.current_block = if_block
        for child in node.body:
            self.visit(child)
        self.add_exit(self.current_block, afterif_block)

        # Continue building the CFG in the after-if block.
        self.current_block = afterif_block

    def visit_While(self, node: While) -> None:
        # TODO while/else

        loop_guard = self.new_loopguard()
        self.current_block = loop_guard
        self.add_statement(self.current_block, node)

        if isinstance(node.test, ast.Call):
            self.visit(node.test)
        # New block for the case where the test in the while is True.
        while_block = self.new_block()
        self.add_exit(self.current_block, while_block, node.test)

        # New block for the case where the test in the while is False.
        afterwhile_block = self.new_block()
        self.add_exit(self.current_block, afterwhile_block, invert(node.test))

        # Populate the while block.
        self.current_block = while_block
        self.loop_stack.appendleft((afterwhile_block, loop_guard))
        for child in node.body:
            self.visit(child)
        top = self.loop_stack.popleft()
        assert top == (afterwhile_block, loop_guard)
        self.add_exit(self.current_block, loop_guard)

        # Continue building the CFG in the after-while block.
        self.current_block = afterwhile_block

    def visit_For(self, node: For) -> None:
        # TODO for/else

        loop_guard = self.new_loopguard()
        self.current_block = loop_guard
        self.add_statement(self.current_block, node)

        if isinstance(node.iter, ast.Call):
            self.visit(node.iter)
        # New block for the body of the for-loop.
        for_block = self.new_block()
        self.add_exit(self.current_block, for_block, node.iter)

        # Block of code after the for loop.
        afterfor_block = self.new_block()
        self.add_exit(self.current_block, afterfor_block)
        self.current_block = for_block

        # Push exit destinations for break/continue statements.
        # On break, control jumps to afterfor_block.
        # On continue, control jumps to loop_guard.
        self.loop_stack.appendleft((afterfor_block, loop_guard))

        # Populate the body of the for loop.
        for child in node.body:
            self.visit(child)

        top = self.loop_stack.popleft()
        assert top == (afterfor_block, loop_guard)
        self.add_exit(self.current_block, loop_guard)

        # Continue building the CFG in the after-for block.
        self.current_block = afterfor_block

    def visit_Break(self, node: Break) -> None:
        assert self.loop_stack
        after_block: Block
        after_block, _ = self.loop_stack[0]
        self.current_block.add_statement(node)
        self.current_block.add_exit(after_block)
        self.current_block = self.new_block()

    def visit_Continue(self, node):
        assert self.loop_stack
        loop_guard: Block
        _, loop_guard = self.loop_stack[0]
        self.add_statement(self.current_block, node)
        self.add_exit(self.current_block, loop_guard)
        self.current_block = self.new_block()

    def visit_Import(self, node: Import) -> None:
        self.add_statement(self.current_block, node)

    def visit_ImportFrom(self, node: ImportFrom) -> None:
        self.add_statement(self.current_block, node)

    def visit_FunctionDef(self, node: FunctionDef) -> None:
        self.add_statement(self.current_block, node)
        self.new_functionCFG(node, asynchr=False)

    def visit_AsyncFunctionDef(self, node):
        self.add_statement(self.current_block, node)
        self.new_functionCFG(node, asynchr=True)

    def visit_Await(self, node):
        afterawait_block = self.new_block()
        self.add_exit(self.current_block, afterawait_block)
        self.generic_visit(node)
        self.current_block = afterawait_block

    def visit_Return(self, node: Return) -> None:
        if self.current_block.statements:
            return_block = self.new_block()
            self.current_block.add_exit(return_block)
            self.current_block = return_block

        self.add_statement(self.current_block, node)

        if self.try_stack:
            stackobj = self.try_stack[0]
            if stackobj.iter_state != TryEnum.FINAL and stackobj.has_final:
                after_return = self.new_block()
                self.current_block.add_exit(after_return)
                after_return.add_exit(self.current_block)
                self.current_block = after_return
                for child in stackobj.node.finalbody:
                    self.visit(child)

        self.cfg.finalblocks.append(self.current_block)
        # Continue in a new block but without any jump to it -> all code after
        # the return statement will not be included in the CFG.
        self.current_block = self.new_block()

    def visit_Yield(self, node: Yield) -> None:
        self.cfg.asynchr = True
        afteryield_block = self.new_block()
        self.add_exit(self.current_block, afteryield_block)
        self.current_block = afteryield_block

    def visit_Try(self, node: Try):
        try_block = self.new_try_block(node)
        after_tryblock = self.new_block()
        self.current_block.add_exit(try_block)
        stackobj = TryStackObject(
            try_block, after_tryblock, bool(node.finalbody)
        )
        self.try_stack.appendleft(stackobj)

        stackobj.iter_state = TryEnum.EXCEPT
        for child in node.handlers:
            self.current_block = self.new_block()
            # If we encounter a raise statement during body iteration,
            # we can link the raise block to the appropriate exception block (if any).
            try:
                try_block.except_blocks[
                    None if child.type is None else child.type.id  # type: ignore
                ] = self.current_block
            except:
                try_block.except_blocks[None] = self.current_block
            self.visit(child)

        stackobj.iter_state = TryEnum.BODY
        self.current_block = try_block
        for child1 in node.body:
            self.visit(child1)

        if node.orelse:
            stackobj.iter_state = TryEnum.ELSE
            else_block = self.new_block()
            self.current_block.add_exit(else_block)
            self.current_block = else_block
            for child2 in node.orelse:
                self.visit(child2)

        self.current_block.add_exit(after_tryblock)
        self.current_block = after_tryblock
        if node.finalbody:
            stackobj.iter_state = TryEnum.FINAL
            for child3 in node.finalbody:
                if isinstance(child3, ast.Raise):
                    top = self.try_stack.popleft()
                    self.visit_Raise(child3)
                    self.try_stack.appendleft(top)
                    break
                self.visit(child3)
            else:
                next_block = self.new_block()
                self.current_block.add_exit(next_block)
                self.current_block = next_block

        del self.try_stack[0]

    def visit_ExceptHandler(self, node: ExceptHandler):
        assert self.try_stack
        for child in node.body:
            self.visit(child)
        self.current_block.add_exit(self.try_stack[0].after_block)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
Just the init!
"""
from . import server
from .builder import CFG, CFGBuilder
import requests
import base64
import os
import pickle
import functools


@functools.singledispatch  # Overload +0
def request_cfg(
    relpath: str, lineno: int, hostname="localhost", port=5000
) -> CFG:
    """
    Request a CFG object from server.
        Overload +0:
            Args:
                relpath (str): Path to python source file relative to server's directory
                lineno: (int): Line number of the desired CFG
            Returns:
                CFG near lineno
        Overload +1:
            Args:
                cfgptr (list): Non-zero length list that stores the CFG
            Returns: A decorator for the desired CFG object
    """
    body = {
        "jsonrpc": "2.0",
        "method": "build_cfg",
        "params": {
            "filename": os.path.join(os.getcwd(), relpath),
            "lineno": lineno,
        },
        "id": 0,
    }

    req = requests.get(f"http://{hostname}:{port}/build", json=body)
    if req.status_code != 200:
        raise Exception(f"Received {req.status_code}")

    res: dict = req.json()["result"]
    if res["error"]:
        raise Exception(f"{res['data']}")
    return pickle.loads(base64.b64decode(res["data"].encode()))


@request_cfg.register  # Overload +1
def _(cfgptr: list, hostname="localhost", port=5000):
    def f(func):
        code = func.__code__
        cfgptr[0] = request_cfg(
            code.co_filename, code.co_firstlineno + 1, hostname, port
        )

        @functools.wraps(func)
        def g(*args, **kwargs):
            return func(*args, **kwargs)

        return g

    return f


__all__ = [
    "server",
    "CFG",
    "CFGBuilder",
    "request_cfg",
]



#! /usr/bin/python3.8
import logging, logging.handlers, logging.config
import os
import subprocess
import argparse
from .util import getcwd

parser = argparse.ArgumentParser()
runner_path = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "_runner.py"
)
parser.add_argument("host", nargs="?", type=str, default="localhost")
parser.add_argument("port", nargs="?", type=str, default="5000")

parser.add_argument("--directory", type=str, default=".")
parser.add_argument(
    "--logconf-port", type=int, default=8123,
)
args = parser.parse_args()
if not os.path.exists(args.directory):
    raise OSError(f"directory {args.directory} does not exist")
else:
    print(f"Running at {os.path.abspath(args.directory)}")
    os.chdir(args.directory)

if not os.path.exists(tmpdir := os.path.join(getcwd(), ".py2cfg")):
    os.mkdir(tmpdir)

logging.config.listen(args.logconf_port).start()
os.environ["FLASK_APP"] = runner_path
subprocess.call(["flask", "run", "-h", args.host, "-p", args.port])



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
This is the wrapper that provides/is the py2cfg shell script.
"""

import os
import sys
import argparse
import pty

# Relative and absolute version of the same thing for interpreter tolerance
sys.path.append("..")
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

# can be either pip or local relative dir
from py2cfg.builder import CFGBuilder  # type: ignore


def main() -> None:
    parser = argparse.ArgumentParser(
        description="Generate the control flow graph of a Python program"
    )
    parser.add_argument(
        "input_file",
        help="Path to a file containing a Python program for which the CFG must be generated",
    )
    parser.add_argument(
        "--short",
        action="store_true",
        help="Shorten all strings above 20 characters",
    )
    parser.add_argument(
        "--format",
        type=str,
        default="svg",
        help="File format of the generated cfg",
    )
    parser.add_argument(
        "--calls", nargs=1, type=bool, default=True, help="Toggle call graph"
    )
    parser.add_argument(
        "--show",
        nargs=1,
        type=bool,
        default=False,
        help="Open CFG after generation",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        default=False,
        help="Generate graphs at run time",
    )
    parser.add_argument(
        "--port", type=int, help="Serve at localhost:port", default=8000,
    )
    parser.add_argument(
        "--wsport", type=int, help="Socket port",
    )
    parser.add_argument(
        "--serve-at",
        type=str,
        help="Host language server here",
        default=os.path.join(os.environ["HOME"]),
    )
    parser.add_argument(
        "--cleanup", action="store_true", help="Remove DOT file"
    )
    parser.add_argument(
        "--pack", type=bool, default=True, help="Generate compact graph"
    )
    parser.add_argument(
        "--level",
        type=str,
        default="WARN",
        choices=["DEBUG", "INFO", "WARN", "CRITICAL", "ERROR",],
    )
    parser.add_argument(
        "--diffable", type=str, help="generate diffable ast traversal"
    )
    args = parser.parse_args()
    os.environ["PY2CFG_SERVER_PORT"] = str(
        8000 if args.port is None else args.port
    )

    os.environ["PY2CFG_SOCKET_PORT"] = str(
        args.port + 1 if args.wsport is None else args.wsport
    )
    os.environ["PY2CFG_PACK"] = str(args.pack)
    os.environ["PY2CFG_SERVE_AT"] = args.serve_at
    os.environ["PY2CFG_CLEANUP"] = str(args.cleanup)
    os.environ["PY2CFG_LOGLEVEL"] = args.level
    cfg_name = args.input_file.split("/")[-1]

    # Run py2cfg interactively along side a debugger.
    if args.debug:
        try:
            from pudb.debugger import Debugger  # type: ignore
            from pudb.run import main as pudb_main  # type: ignore
        except:
            # Once we have a working implementation creating a Spyder plugin should be ez-pz.
            print("pudb not found :(")
            exit(1)

        if not os.path.exists(args.input_file):
            raise OSError("File not found")
        os.environ["PY2CFG_SERVE_AT"] = os.path.join(args.serve_at, ".py2cfg")
        if not os.path.exists(args.serve_at):
            raise OSError("Server directory does not exist")

        if (child := pty.fork()[0]) == 0:
            # Create an http server that serves control flow graph files. Whenever
            # the program is paused, generate a new py2cfg file and notify the
            # browser script.
            from py2cfg._serve import serve

            serve(args.port)
        else:
            from py2cfg._debug_hooks import debug_init
            from py2cfg._serve import set_pid

            set_pid(child)  # child is deleted from global namespace later
            debug_init(Debugger, args.input_file, args.port)
            # What to do about command line args? Use a positional argument as
            # as a toggle?

            # Just create a graph for examples/fib.py
            # $ py2cfg graph [FILE] [--show {True, False}]...

            # Run in debug mode
            # $ py2cfg debug examples/fib.py [--arg [FOO]]
            pudb_main()
            # pudb wipes global variables so we have to import everything
            # without upsetting the parser.
            from os import kill
            from os import getpid
            from signal import SIGINT, SIGTERM
            from py2cfg._serve import get_pid
            from py2cfg._debug_hooks import get_future_id

            if get_future_id() is not None:
                kill(get_future_id(), SIGTERM)
            kill(get_pid(), SIGINT)
            return

    cfg = CFGBuilder(args.short).build_from_file(cfg_name, args.input_file)

    # Some options for wrapping:
    # cfg.build_visual(cfg_name[:-3] + '_cfg', format='pdf', calls=True)
    # cfg.build_visual(cfg_name[:-3] + "_cfg", format="svg", calls=True, show=False)
    # cfg.build_visual(cfg_name[:-3] + "_cfg", format="png", calls=True, show=False)
    # cfg.build_visual('controlflowgraph', format='png', calls=True, show=False)
    cfg.build_visual(
        cfg_name[:-3] + "_cfg",
        format=args.format,
        calls=args.calls,
        show=args.show,
        diffable=args.diffable,
    )


if __name__ == "__main__":
    main()



import logging
import os
from watchdog.observers import Observer
from watchdog.events import FileSystemEvent, RegexMatchingEventHandler
from py2cfg.builder import CFGBuilder
from .util import getcwd, normalizer
from .db import connect, process_source_file, delete_rows

logger = logging.getLogger(__name__)


class FileHandler(RegexMatchingEventHandler):
    def on_modified(self, event: FileSystemEvent):
        src_path = normalizer(event.src_path)
        with connect() as cursor:
            try:
                cfg = CFGBuilder(short=True).build_from_file("", src_path)
            except:
                return
            else:
                delete_rows(src_path, cursor)
                process_source_file(cfg, src_path, cursor)
                logger.warn(f"Modified: {src_path}")

    def on_deleted(self, event):
        src_path = normalizer(event.src_path)
        with connect() as cursor:
            delete_rows(src_path, cursor)
            logger.warn(f"Deleted: {src_path}")


def create_managed_database(
    ignore_regexes=None, ignore_directories=False, case_sensitive=False,
) -> None:
    observer = Observer()
    observer.schedule(
        FileHandler(
            regexes=[r".*\.py$"],
            ignore_regexes=ignore_regexes,
            ignore_directories=ignore_directories,
            case_sensitive=case_sensitive,
        ),
        getcwd(),
        recursive=True,
    )
    observer.start()

    with connect() as cursor:
        for root, path, files in os.walk(getcwd()):
            for file in files:
                if file.endswith(".py"):
                    filepath = os.path.join(root, file)
                    try:
                        cfg = CFGBuilder(short=True).build_from_file(
                            "", filepath
                        )
                    except:
                        logger.error(f"Failed to build cfg {filepath}")
                        continue
                    else:
                        process_source_file(cfg, filepath, cursor)
                        logger.debug(f"Built {filepath}")



import logging
import os
import contextlib
import http.server
import shutil


logger = logging.getLogger(__name__)
logger.setLevel(1)

SUBDIR = ".py2cfg"
PROTODIR = "_proto_"


@contextlib.contextmanager
def serve_dir(serve_dir):
    # Need a place to host generated control flow graphs/metadata/whatever.

    serve_path = os.path.join(serve_dir, SUBDIR)
    proto = os.path.join(os.path.dirname(__file__), PROTODIR)
    logger.debug(f"Copying prototype at {proto} to {serve_path}")

    if os.path.exists(serve_path):
        shutil.rmtree(serve_path)

    try:
        shutil.copytree(proto, serve_path)
    except OSError:
        logger.exception(f"Failed to copy tree")
        return -1
    try:
        yield serve_path
    finally:
        shutil.rmtree(serve_path)


_pid = None


def set_pid(pid):
    global _pid
    _pid = pid


def get_pid():
    return _pid


def serve(port=None, directory=None, pipe=None) -> None:
    """Serve control flow graph files"""
    rootlog = logging.getLogger()
    rootlog.addHandler((fh := logging.FileHandler("log.txt")))
    fh.setFormatter(logging.Formatter("%(name)s - %(levelname)s - %(message)s"))

    port = 8000 if port is None else port
    directory = os.getenv("HOME") if directory is None else directory

    with serve_dir(directory) as target:

        class Py2CFG_Handler(http.server.SimpleHTTPRequestHandler):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, directory=target, **kwargs)

            # Only works for some browsers. Mangling filename still required
            def end_headers(self):
                # https://gist.github.com/dustingetz/5348582
                self.send_header(
                    "Cache-Control", "no-cache, no-store, must-revalidate",
                )
                self.send_header("Pragma", "no-cache")
                self.send_header("Expires", "0")
                super().end_headers()

        with http.server.HTTPServer(("", port), Py2CFG_Handler) as httpd:
            httpd.serve_forever()



from py2cfg.model import Block
import sqlite3
import threading
import os
from typing import List, Tuple
from py2cfg.builder import CFG
import pickle
import logging
import contextlib
import time
from .util import getcwd

logger = logging.getLogger(__name__)


TMPDIR = os.path.join(getcwd(), ".py2cfg")
DB_FILE = os.path.join(getcwd(), ".py2cfg", "cfg.db")
UPDATE_LOCK = threading.Condition()


def _sql_init():
    if not os.path.exists(DB_FILE):
        # Server initiated in new directory so define schema (?)
        cursor = sqlite3.connect(os.path.join(TMPDIR, "cfg.db")).cursor()
        cursor.execute(
            r"""
            CREATE TABLE Files (
                qualname TEXT,
                filepath TEXT,
                lineno INT,
                end_lineno INT,
                PRIMARY KEY (qualname),
                FOREIGN KEY (filepath) REFERENCES Data(filepath)  
            )
            """,
        )
        cursor.execute(
            r"""
                CREATE TABLE Data (
                    filepath TEXT NOT NULL,  
                    data BLOB,
                    PRIMARY KEY (filepath)
                )
            """,
        )

    return sqlite3.connect(
        os.path.join(TMPDIR, "cfg.db"),
        isolation_level="IMMEDIATE",
        cached_statements=0,
    )


@contextlib.contextmanager
def connect() -> sqlite3.Cursor:
    with UPDATE_LOCK:
        conn = _sql_init()
        try:
            yield conn.cursor()
        finally:
            conn.commit()
            # It seems that sqlite is not MT-Safe so wait for all transactions
            # to finalize.
            while conn.in_transaction:
                time.sleep(0.01)
            conn.close()


sortblocks = lambda cfg: list(sorted(cfg.own_blocks(), key=lambda x: x.at()))


def process_source_file(cfg: CFG, filepath, cursor: sqlite3.Cursor):
    fields = set()

    def set_qualname(cfg: CFG, qualname: str):
        classgraph: CFG
        funcgraph: CFG
        for classgraph in cfg.classcfgs.values():
            classgraph.qualname = f"{qualname}.{classgraph.name}"
            set_qualname(classgraph, classgraph.qualname)
            fields.add(
                (classgraph.qualname, classgraph.lineno, classgraph.end_lineno,)
            )

        for funcgraph in cfg.functioncfgs.values():
            funcgraph.qualname = f"{qualname}.{funcgraph.name}"
            set_qualname(funcgraph, funcgraph.qualname)
            fields.add(
                (funcgraph.qualname, funcgraph.lineno, funcgraph.end_lineno,)
            )

    relpath = os.path.relpath(filepath)
    cfg.qualname = f"<{relpath}>"
    set_qualname(cfg, f"<{relpath}>")
    fields.add((f"<{relpath}>", cfg.lineno, cfg.end_lineno,))
    for qualname, lineno, end_lineno in fields:
        cursor.execute(
            r"""
                INSERT INTO Files (
                    qualname,
                    filepath,
                    lineno,
                    end_lineno
                ) VALUES (
                    ?, ?, ?, ?
                ) ON CONFLICT(qualname) DO
                UPDATE SET lineno=?, end_lineno=?, filepath=?
            """,
            [
                qualname,
                lineno,
                end_lineno,
                filepath,
                lineno,
                end_lineno,
                filepath,
            ],
        )
    data = pickle.dumps(cfg)
    cursor.execute(
        r"""
            INSERT INTO Data (
                filepath,
                data
            ) Values (
                ?, ?
            ) ON CONFLICT(filepath) DO
            UPDATE SET data=?
        """,
        [filepath, data, data],
    )

    for (qualname, lineno, end_lineno,) in fields:
        print_this = "{a:50}, {b:>3}, {c:>3},".format(
            a=qualname, b=lineno, c=end_lineno,
        )
        logger.debug(print_this)


def query_filename_lineno(
    filename: str, lineno: int, cursor: sqlite3.Cursor,
) -> Tuple[CFG, List[Block]]:

    cursor.execute(
        r"""
            SELECT files.qualname, data.data, files.lineno, files.end_lineno
            FROM files
            INNER JOIN data ON files.filepath=data.filepath
            WHERE files.filepath=? AND
                lineno<=? AND
                end_lineno>=?
    """,
        (filename, lineno, lineno),
    )

    # Which enclosing scope is the mostest-closest?
    def distance(start, end, lineno):
        start_dist, end_dist = abs(lineno - start), abs(lineno - end)
        return (start_dist + end_dist) / 2

    qualname: str
    cfg: CFG
    selected = cursor.fetchall()
    if not selected:
        raise Exception(f"{filename} not found in database")

    qualname, blob, _, _ = min(
        selected, key=lambda item: distance(item[2], item[3], lineno)
    )
    cfg = pickle.loads(blob)  # Possibly dangerous
    module_qual = f"<{os.path.relpath(filename)}>"
    qualname = qualname.replace(module_qual, "")
    if not qualname:
        return cfg, sortblocks(cfg)

    # Traverse path until we find the mostest-closest enclosing scope
    subcfg_path = qualname.split(".")[1:]
    while subcfg_path:
        name = subcfg_path.pop(0)
        if name in cfg.classcfgs:
            cfg = cfg.classcfgs[name]
        elif name in cfg.functioncfgs:
            cfg = cfg.functioncfgs[name]
        else:
            raise Exception("Reached supposedly unreachable code")

    return cfg, sortblocks(cfg)


def delete_rows(filepath, cursor: sqlite3.Cursor):
    cursor.execute(
        r"""
        DELETE FROM data
        WHERE filepath=?
    """,
        [filepath],
    )
    cursor.execute(
        r"""
        DELETE FROM files
        WHERE filepath=?
    """,
        [filepath],
    )



from collections import defaultdict
from py2cfg.model import Block, CFG
from typing import Dict, Tuple, List


class _Database:
    def __init__(self):
        self.__dict__ = defaultdict(dict)


F_lineno = Tuple[str, int]  # filename, lineno
F_function = Tuple[str, str]  # filename, funcname
CFG_Blocks = Tuple[CFG, List[Block]]


class CFG_Database(_Database):
    """Database of CFG Indices"""

    def has_file(self, filename):
        return filename in set(f[0] for f in self.__dict__.keys())

    def index(self, filename, cfg: CFG):
        assert not self.has_file(filename)
        sortblocks = lambda cfg: list(
            sorted(cfg.own_blocks(), key=lambda x: x.at())
        )
        topcfg = cfg, sortblocks(cfg)
        if not cfg.entryblock:
            raise TypeError("Expecting cfg.entryblock to be not None")
        self.at_lineno[filename, cfg.entryblock.at()] = topcfg
        self.at_function[filename, "<module>"] = topcfg

        for funcgraph in cfg.functioncfgs.values():
            subcfg = funcgraph, sortblocks(funcgraph)
            self.at_lineno[filename, funcgraph.lineno] = subcfg
            self.at_function[filename, funcgraph.name] = subcfg

        for classgraph in cfg.classcfgs.values():
            pass

    @property
    def at_lineno(self) -> Dict[F_lineno, CFG_Blocks]:
        """A mapping indexed by lineno of function def"""
        return self.__dict__["_lineno"]  # type: ignore

    @property
    def at_function(self) -> Dict[F_function, CFG_Blocks]:
        """A mapping indexed by name of function"""
        return self.__dict__["_function"]  # type: ignore



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
Control flow graph for Python programs.
"""

import collections
from typing import Any, Deque, Tuple, List, Optional, Iterator, Set, Dict
import ast  # type: ignore
from _ast import Compare  # For type-hinting
import astor  # type: ignore
import graphviz as gv  # type: ignore
from graphviz.dot import Digraph  # type: ignore
import re
from collections import Counter, deque
import os


class Block(object):
    """
    Basic block in a control flow graph.

    Contains a list of statements executed in a program without any control
    jumps. A block of statements is exited through one of its exits. Exits are
    a list of Links that represent control flow jumps.
    """

    __slots__ = (
        "id",
        "statements",
        "func_calls",
        "predecessors",
        "exits",
        "func_blocks",
        "highlight",
        "outline",
    )

    def __init__(self, id: int) -> None:
        # Id of the block.
        self.id = id
        # Statements in the block.
        self.statements: List[ast.AST] = []
        # Calls to functions inside the block (represents context switches to
        # some functions' CFGs).
        self.func_calls: List[str] = []
        # Links to predecessors in a control flow graph.
        self.predecessors: List[Link] = []
        # Links to the next blocks in a control flow graph.
        self.exits: List[Link] = []
        # Function blocks within this block
        self.func_blocks: List[FuncBlock] = []
        # Interactive styling
        self.highlight: bool = False
        # Debugger at this block
        self.outline: bool = False

    def __iter__(self):
        return iter(self.statements)

    def __str__(self) -> str:
        if self.statements:
            return "block:{}@{}".format(self.id, self.at())
        return "empty block:{}".format(self.id)

    def __repr__(self) -> str:
        txt = "{} with {} exits".format(str(self), len(self.exits))
        if self.statements:
            txt += ", body=["
            txt += ", ".join([ast.dump(node) for node in self.statements])
            txt += "]"
        return txt

    def at(self) -> int:
        """
        Get the line number of the first statement of the block in the program.
        """
        if self.statements and self.statements[0].lineno >= 0:
            return self.statements[0].lineno
        return -1

    def end(self) -> int:
        """
        Get the line number of the last statement of the block in the program.
        """
        if self.statements and self.statements[-1].lineno >= 0:
            return self.statements[-1].lineno
        return -1

    def is_empty(self) -> bool:
        """
        Check if the block is empty.

        Returns:
            A boolean indicating if the block is empty (True) or not (False).
        """
        return not self.statements

    def get_source(self) -> str:
        """
        Get a string containing the Python source code corresponding to the
        statements in the block.

        Returns:
            A string containing the source code of the statements.
        """
        src = ""
        for statement in self.statements:
            if type(statement) in [ast.If, ast.For, ast.While]:
                src += astor.to_source(statement).split("\n")[0] + "\n"
            elif (
                type(statement) == ast.FunctionDef
                or type(statement) == ast.AsyncFunctionDef
            ):
                src += (astor.to_source(statement)).split("\n")[0] + "...\n"
            else:
                src += astor.to_source(statement)
        return src

    def get_calls(self) -> str:
        """
        Get a string containing the calls to other functions inside the block.

        Returns:
            A string containing the names of the functions called inside the
            block.
        """
        txt = ""
        for func_name in list(set(self.func_calls)):
            txt += func_name + "\n"
        return txt

    def type(self, default: Optional[Any] = None) -> Any:
        default = ast.AST if default is None else default
        return type(self.statements[0]) if self.statements else default

    def __hash__(self) -> int:
        # Allows Block objects to be used as dict keys and set elements
        # return id(self)
        return self.id

    def add_statement(self, node: ast.AST):
        self.statements.append(node)

    def add_exit(self, next, exitcase=None) -> None:
        link = Link(self, next, exitcase)
        self.exits.append(link)
        next.predecessors.append(link)


class Link(object):
    """
    Link between blocks in a control flow graph.

    Represents a control flow jump between two blocks. Contains an exitcase in
    the form of an expression, representing the case in which the associated
    control jump is made.
    """

    __slots__ = (
        "source",
        "target",
        "exitcase",
        "highlight",
    )

    def __init__(
        self,
        source: Block,
        target: Block,
        exitcase: Optional[ast.Compare] = None,
    ) -> None:
        assert isinstance(target, Block), "Source of a link must be a block"
        assert isinstance(target, Block), "Target of a link must be a block"
        # Block from which the control flow jump was made.
        self.source = source
        # Target block of the control flow jump.
        self.target = target
        # 'Case' leading to a control flow jump through this link.
        self.exitcase = exitcase
        # Interactive styling
        self.highlight: bool = False

    def __str__(self) -> str:
        return f"link from {self.source} to {self.target}"

    def __repr__(self) -> str:
        # This isn't how repr is supposed to be used... We should be able to
        # deep copy this object by calling eval(repr(link))`.
        if self.exitcase is not None:
            return f"{self}, with exitcase {ast.dump(self.exitcase)}"
        return str(self)

    def jumpfrom(self) -> Optional[int]:
        """Return the line of source end"""
        return self.source.end()

    def jumpto(self) -> Optional[int]:
        """Return the line of target start"""
        return self.target.at()

    def get_exitcase(self) -> str:
        """
        Get a string containing the Python source code corresponding to the
        exitcase of the Link.

        Returns:
            A string containing the source code.
        """
        if self.exitcase:
            return astor.to_source(self.exitcase)
        return ""


class FuncBlock(Block):
    __slots__ = ("args", "name")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.args: List[FuncBlock] = []
        self.name: Optional[str] = None


class TryBlock(Block):
    __slots__ = ("except_blocks",)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.except_blocks: Dict[Optional[Exception], Block] = {}

    def get_source(self) -> str:
        """
        Get a string containing the Python source code corresponding to the
        statements in the block.

        Returns:
            A string containing the source code of the statements.
        """
        if not self.statements[1:]:
            return "try:"
        src = ""
        for statement in self.statements[1:]:  # We just want try body
            if type(statement) in [ast.If, ast.For, ast.While]:
                src += astor.to_source(statement).split("\n")[0] + "\n"
            elif (
                type(statement) == ast.FunctionDef
                or type(statement) == ast.AsyncFunctionDef
            ):
                src += (astor.to_source(statement)).split("\n")[0] + "...\n"
            else:
                src += astor.to_source(statement)
        return src


class CFG(object):
    """
    Control flow graph (CFG).

    A control flow graph is composed of basic blocks and links between them
    representing control flow jumps. It has a unique entry block and several
    possible 'final' blocks (blocks with no exits representing the end of the
    CFG).
    """

    # Also serves as graph Key table
    # TODO Change value type to dict. Can be upacked into graph.node fn.
    node_styles = {
        "input": ("parallelogram", "#afeeee"),  # Pale Turquoise
        "default": ("rectangle", "#FFFB81"),  # Pale Yellow
        ast.If: ("diamond", "#FF6752"),  # Pale Red
        ast.For: ("hexagon", "#FFBE52"),  # Pale Orange
        ast.While: ("hexagon", "#FFBE52"),  # Pale Orange
        ast.Call: ("tab", "#E552FF"),  # Pale Purple
        ast.Return: ("parallelogram", "#98fb98"),  # Pale Green
        ast.Try: ("Mdiamond", "orange"),
        ast.Raise: ("house", "#98fb98"),
    }
    DEFAULT = node_styles["default"]

    def __init__(
        self, name: str, asynchr: bool = False, short: bool = True
    ) -> None:
        assert isinstance(name, str), "Name of a CFG must be a string"
        assert isinstance(asynchr, bool), "Async must be a boolean value"
        # Name of the function or module being represented.
        self.name = name
        # Type of function represented by the CFG (sync or async). A Python
        # program is considered as a synchronous function (main).
        self.asynchr = asynchr
        # Entry block of the CFG.
        self.entryblock: Optional[Block] = None
        # Final blocks of the CFG.
        self.finalblocks: List[Block] = []
        # Sub-CFGs for functions defined inside the current CFG.
        self.functioncfgs: dict = {}
        # Sub-CFGs
        self.classcfgs: dict = {}
        self.isShort = short
        self.lineno: int = 0
        self.end_lineno: int = 0
        self.qualname = ""

    def __str__(self) -> str:
        return "CFG for {}".format(self.name)

    def _build_key_subgraph(self, format: str) -> Digraph:
        # Generates a key for th?!?jedi=0, e CFG?!? (name=None, comment=None, filename=None, directory=None, format=None, engine=None, encoding=backend.ENCODING, graph_attr=None, node_attr=None, *_*edge_attr=None*_*, body=None, strict=False) ?!?jedi?!?
        key_subgraph = gv.Digraph(
            name="cluster_KEY",
            format=format,
            graph_attr={"label": "KEY", "fontname": "DejaVu Sans Mono"},
            node_attr={"fontname": "DejaVu Sans Mono"},
            edge_attr={"fontname": "DejaVu Sans Mono"},
        )
        for typeobj, (shape, color) in zip(
            self.node_styles.keys(), self.node_styles.values()
        ):
            key_subgraph.node(
                name=(
                    typeobj.__name__.lower()
                    if isinstance(typeobj, type)
                    else str(typeobj)
                ),
                _attributes={
                    "shape": shape,
                    "style": "filled",
                    "fillcolor": color,
                },
            )

        # Adding edge helps with alignment.
        key_subgraph.edge("if", "input", _attributes={"style": "invis"})
        key_subgraph.edge("input", "call", _attributes={"style": "invis"})
        key_subgraph.edge("for", "return", _attributes={"style": "invis"})
        key_subgraph.edge("return", "default", _attributes={"style": "invis"})
        key_subgraph.edge("try", "raise", _attributes={"style": "invis"})
        return key_subgraph

    # default handler for graph nodes
    def _style_handler(
        self, block: Block, typeobject: Any
    ) -> Tuple[str, str, str]:

        shape, color = self.node_styles.get(typeobject, self.DEFAULT)
        # Left-align all remaining blocks
        text = "\l".join(line for line in block.get_source().splitlines())
        return shape, color, text

    # default handler for graph edges
    def _edge_handler(self, link: Link) -> Tuple[None, str, str]:
        return None, "black", link.get_exitcase().strip()

    def stylize_node(self, block: Block, default: None = None) -> tuple:
        # Since we're modelling ast objects, we might as well copy how they
        # implement handlers. This should allow us to implement new node types
        # simply defining a method `node_ASTNAME`
        typeobject = block.type(default)
        name = (
            typeobject.__name__ if isinstance(typeobject, type) else typeobject
        )
        style = getattr(self, f"node_{name}", self._style_handler)(
            block, typeobject
        )
        assert (
            len(style) == 3
        ), "Style handler must return (shape, color, label)"
        assert all(
            isinstance(attr, str) or attr or attr is None for attr in style
        ), "Style attributes must be str or NoneType"
        return style

    def stylize_edge(
        self, link: Link, default: None = None, default_target: None = None
    ) -> tuple:
        typeobject = link.source.type(default)
        name = (
            typeobject.__name__ if isinstance(typeobject, type) else typeobject
        )
        target_type = link.target.type(default_target)
        target_name = (
            target_type.__name__ if isinstance(typeobject, type) else typeobject
        )
        # Do specific attribute search of source->dest
        style_fn = getattr(self, f"{name}_to_{target_name}", None)

        # Do generic attribute search of source->any
        if style_fn is None:
            style_fn = getattr(self, f"edge_{name}", self._edge_handler)
        style = style_fn(link)
        assert (
            len(style) == 3
        ), "style handler must return (shape, color, label)"
        assert all(
            isinstance(attr, str) or attr is None for attr in style
        ), "style attributes must be str or NoneType"
        return style

    def _visit_func(
        self,
        graph: Digraph,
        block: FuncBlock,
        last: str = None,
        visited: Set[Block] = set(),
        interactive: bool = False,
    ) -> None:
        if block in visited:
            return
        visited.add(block)
        shape, color, _ = self.stylize_node(block)
        graph.node(
            str(block.id),
            label=block.name,
            _attributes={"shape": shape, "color": color, "style": "filled"},
        )
        if last is not None:
            graph.edge(last, str(block.id), _attributes={"color": "black"})
        for arg in block.args:
            self._visit_func(graph, arg, str(block.id), interactive=interactive)

    def _visit_blocks(
        self,
        graph: Digraph,
        block: Block,
        visited: Set[Block] = set(),
        calls: bool = True,
        format: str = None,
        interactive: bool = False,
    ) -> None:
        # Don't visit blocks twice.
        if block in visited:
            return
        visited.add(block)
        nodeshape, nodecolor, nodelabel = self.stylize_node(block)
        node_type = block.type()
        original_nodelabel = nodelabel
        nodelabel = ""
        if (
            self.isShort
            and not isinstance(node_type, ast.ClassDef)
            and not isinstance(node_type, ast.FunctionDef)
            and not isinstance(node_type, ast.If)
            and not isinstance(node_type, ast.While)
        ):
            sub_pattern = r"""(\"|')                # Group 1: " or '
                              (?=[^\"'\r\n]{20,})   # Enforce min length of 20
                              ([^\"'\r\n]{,20})     # Group 2: Words that stay
                              ([^\"'\r\n]{,9999})   # Group 3: Shorten these 
                              (\"|')"""  # Group 4: " or '
            original_nodelabel = original_nodelabel.replace("\l", "\n")
            for line in original_nodelabel.splitlines():
                tmp_line = re.sub(
                    sub_pattern, r"\1\2...\4", line, flags=re.VERBOSE
                )
                nodelabel += tmp_line + "\l"
        else:
            nodelabel = original_nodelabel

        graph.node(
            str(block.id),
            label=nodelabel,
            _attributes={
                "style": f"filled,{self.border_style(block, interactive)}",
                "shape": nodeshape,
                "fillcolor": self.fillcolor(block, interactive, nodecolor),
            },
        )

        if isinstance(block, TryBlock):
            for except_block in block.except_blocks.values():
                self._visit_blocks(graph, except_block, visited, calls, format)

        if calls and block.func_calls:
            calls_node = str(block.id) + "_calls"

            # Remove any duplicates by splitting on newlines and creating a set
            calls_label = block.get_calls().strip()

            # Create a new subgraph for call statement
            calls_subgraph = gv.Digraph(
                name=f"cluster_{block.id}",
                format=format,
                graph_attr={
                    "rankdir": "TB",
                    "ranksep": "0.02",
                    "style": "filled",
                    "color": self.fillcolor(block, interactive, "purple"),
                    "compound": "true",
                    "fontname": "DejaVu Sans Mono",
                    "shape": self.node_styles[ast.Call][0],
                    "label": "",
                },
                node_attr={"fontname": "DejaVu Sans Mono"},
                edge_attr={"fontname": "DejaVu Sans Mono"},
            )
            # Generate control flow edges for function arguments
            for func_block in block.func_blocks:
                graph.edge(
                    str(block.id),
                    str(func_block.id),
                    label="calls",
                    _attributes={"style": "dashed"},
                )
                self._visit_func(
                    calls_subgraph,
                    func_block,
                    visited=set(),
                    interactive=interactive,
                )
            graph.subgraph(calls_subgraph)

            tmp = ""
            for line in calls_label.splitlines():
                if "input" in line:
                    input_node = str(block.id) + "_input"
                    nodeshape, nodecolor = self.node_styles["input"]
                    graph.node(
                        input_node,
                        label=line,
                        _attributes={
                            "style": f"filled,{self.border_style(block, interactive)}",
                            "shape": nodeshape,
                            "fillcolor": self.fillcolor(
                                block, interactive, nodecolor
                            ),
                        },
                    )
                    graph.edge(input_node, str(block.id))  # yellow
                    # _attributes={'style': 'dashed'})

                else:
                    line += "\l"
                    tmp += line

        # Recursively visit all the blocks of the CFG.
        for exit in block.exits:
            assert block == exit.source
            self._visit_blocks(
                graph,
                exit.target,
                visited,
                calls=calls,
                format=format,
                interactive=interactive,
            )
            edgeshape, edgecolor, edgelabel = self.stylize_edge(exit)
            graph.edge(
                str(block.id),
                str(exit.target.id),
                label=edgelabel,
                _attributes={"color": edgecolor},
            )

    subgraphs: dict = Counter()

    def _build_visual(
        self,
        format: str = "pdf",
        calls: bool = True,
        interactive: bool = False,
        build_own=False,
    ) -> Digraph:
        graph = gv.Digraph(
            name=f"cluster{self.subgraphs[self.name]}{self.name}",
            format=format,
            graph_attr={
                "label": self.name,
                "rankdir": "TB",
                "ranksep": "0.02",
                "fontname": "DejaVu Sans Mono",
                "compound": "True",
                "pack": "False",
            },
            node_attr={"fontname": "DejaVu Sans Mono"},
            edge_attr={"fontname": "DejaVu Sans Mono"},
        )
        self.subgraphs[self.name] += 1
        if self.entryblock is None:
            raise TypeError(
                "Expected self.entryblock to be not None but type is None"
            )
        self._visit_blocks(
            graph,
            self.entryblock,
            visited=set(),
            calls=calls,
            format=format,
            interactive=interactive,
        )
        if build_own:
            return graph
        # Build the subgraphs for the function definitions in the CFG and add
        # them to the graph.
        for subcfg in self.classcfgs:
            subgraph = self.classcfgs[subcfg]._build_visual(
                format=format, calls=calls, interactive=interactive,
            )
            graph.subgraph(subgraph)

        for subcfg in self.functioncfgs:
            subgraph = self.functioncfgs[subcfg]._build_visual(
                format=format, calls=calls, interactive=interactive,
            )
            graph.subgraph(subgraph)
        return graph

    @staticmethod
    def border_style(block: Block, interactive: bool):
        if interactive:
            return "bold" if block.outline else "dashed"
        else:
            return "solid"

    @staticmethod
    def fillcolor(
        block: Block, interactive: bool, color: str, default: str = ""
    ):
        if interactive:
            if block.highlight:
                return color
            else:
                return default
        else:
            return color

    def build_key(self, **kwargs):
        """Build key graph"""
        key = self._build_key_subgraph(kwargs.get("format", "svg"))
        return key.render(**kwargs)

    def build_visual(
        self,
        filepath: str,
        format: str,
        calls: bool = True,
        show: bool = True,
        cleanup=True,
        directory=None,
        interactive=False,
        build_keys=True,
        build_own=False,
        diffable=None,
    ) -> str:
        """
        Build a visualisation of the CFG with graphviz and output it in a DOT
        file.

        Args:
            filename: The name of the output file in which the visualisation
                      must be saved.
            format: The format to use for the output file (PDF, ...).
            show: A boolean indicating whether to automatically open the output
                  file after building the visualisation.
        """

        if diffable is not None:
            with open(diffable, "w") as fp:
                for block in self:
                    print(
                        f"{block.id} {type(block.statements[0]).__name__}",
                        file=fp,
                    )

        graph = self._build_visual(format, calls, interactive, build_own)
        if build_keys:
            graph.subgraph(self._build_key_subgraph(format))
        return graph.render(
            filepath, view=show, cleanup=cleanup, directory=directory
        )

    def __iter__(self) -> Iterator[Block]:
        """
        Generator that yields all the blocks in the current graph, then
        recursively yields from any sub graphs
        """
        yield from self.own_blocks()
        for subcfg in self.classcfgs.values():
            yield from subcfg

        for subcfg in self.functioncfgs.values():
            yield from subcfg

    def own_blocks(self) -> Iterator[Block]:
        """
        Generator that yields all blocks in the current graph, excluding any
        subgraphs
        """
        visited = set()
        if self.entryblock is None:
            raise TypeError(
                "Expected self.entryblock to be not None but type is None"
            )
        to_visit: Deque[Block] = deque([self.entryblock])
        while to_visit:
            block = to_visit.popleft()
            visited.add(block)
            for exit_ in block.exits:
                if exit_.target in visited or exit_.target in to_visit:
                    continue
                to_visit.append(exit_.target)
            yield block

    def node_ClassDef(
        self, block: Block, typeobj: type
    ) -> Tuple[str, str, str]:
        text = block.get_source().splitlines()[0] + "...\n"
        return (*self.node_styles.get(typeobj, self.DEFAULT), text)

    def node_Try(self, block: Block, typeobj: type) -> Tuple[str, str, str]:
        return (
            *self.node_styles.get(typeobj, self.DEFAULT),
            "\n".join(block.get_source().splitlines()),
        )

    def edge_If(self, link: Link) -> Tuple[None, str, str]:
        _, _, nodelabel = self.stylize_node(link.source)
        color = "red"
        edgelabel = link.get_exitcase().strip()
        if edgelabel in nodelabel:
            color = "green"
        return None, color, edgelabel

    def edge_While(self, link: Link) -> Tuple[None, str, str]:
        _, _, nodelabel = self.stylize_node(link.source)
        color = "red"
        edgelabel = link.get_exitcase().strip()
        if edgelabel in nodelabel:
            color = "green"
        return None, color, edgelabel

    def edge_For(self, link: Link) -> Tuple[None, str, str]:
        _, _, nodelabel = self.stylize_node(link.source)
        color = "red"
        edgelabel = link.get_exitcase().strip()
        if edgelabel in nodelabel:
            color = "green"
        return None, color, edgelabel

    def bsearch(
        self, lineno: int, lst: Optional[list] = None
    ) -> Optional[Block]:
        """Search for a block at line"""

        def _bsearch(lst, low, high, line):
            if high >= low:
                mid = (low + high) // 2
                block = lst[mid]
                if block.at() <= line <= block.end():
                    return block
                elif line < block.at():
                    return _bsearch(lst, low, mid - 1, line)
                else:
                    return _bsearch(lst, mid + 1, high, line)
            else:
                return None

        lst = list(self) if lst is None else lst  # Already sorted by lineno
        return _bsearch(lst, 0, len(lst) - 1, lineno)

    outlined_block = Block(0)
    highlighted_blocks: List[Block] = []

    def outline_block(self, lineno, lst=None) -> int:
        """In interactive mode, outline the block at lineno"""
        self.outlined_block.outline = False
        if (block := self.bsearch(lineno, lst)) is None:
            return -1
        self.outlined_block = block  # type: ignore
        block.outline = True  # type: ignore
        return 0

    def highlight_blocks(self, lines, blocks=None) -> int:
        """In interactive mode, highlight the block at lineno"""
        for block in self.highlighted_blocks:
            block.highlight = False

        highlighted_blocks: List[Block] = []
        for line in list(lines):
            if (block := self.bsearch(line, blocks)) is not None:  # type: ignore
                block.highlight = True
                highlighted_blocks.append(block)

        self.highlighted_blocks = highlighted_blocks
        return 0

    def find_path(self, finalblock: Block) -> Deque[Link]:
        if self.entryblock is None:
            raise ValueError("entryblock cannot be none")

        assert finalblock in self.finalblocks
        visited: Set[Block] = set()
        path: Deque[Link] = collections.deque()

        def _find_path(link: Link):
            blk: Block = link.target
            if blk in visited:
                return False

            visited.add(blk)
            if blk == finalblock:
                return True

            for lk in blk.exits:
                path.append(lk)
                if _find_path(lk):
                    return True
                path.pop()

        for lk in self.entryblock.exits:
            path.append(lk)
            if _find_path(lk):
                return path
            path.pop()

        # TODO find all possible paths from entry to final and return set
        # TODO find path from any line to any line
        return collections.deque()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['_proto_',
 '_proto_.src',
 'py2cfg',
 'py2cfg._proto_',
 'py2cfg._proto_.src',
 'py2cfg.server',
 'server',
 'src']

package_data = \
{'': ['*']}

install_requires = \
['astor>=0.8.1,<0.9.0',
 'flask>=1.1.2,<2.0.0',
 'graphviz>=0.16,<0.17',
 'pudb>=2020.1,<2021.0',
 'requests>=2.25.1,<3.0.0',
 'websocket-server>=0.4,<0.5']

entry_points = \
{'console_scripts': ['py2cfg = py2cfg._runner:main']}

setup_kwargs = {
    'name': 'py2cfg',
    'version': '0.7.1',
    'description': 'Converts python source code to colorful Control Flow Graphs (CFGs).',
    'long_description': "# py2cfg\nPython3 control flow graph generator\n\n`py2cfg` is a package that can be used to produce control flow graphs (CFGs) for Python 3 programs. \nThe CFGs it generates can be easily visualised with graphviz.\nThat graphical analysis is the main purpose of the module.\n\n## Examples\nBelow is an example of a piece of code that generates the Fibonacci sequence and the CFG produced for it with py2cfg:\n\n```py\n# fib.py\n\ndef fib():\n    a, b = 0, 1\n    while True:\n        yield a\n        a, b = b, a + b\n\nfib_gen = fib()\nfor _ in range(10):\n    next(fib_gen)\n```\n\n![](_static/fib_cfg.svg)\n![](docs/source/_static/fib_cfg.svg)\n\n```py\nclass Sort:\n    def merge(self, l, r):\n        i = 0\n        j = 0\n        arr = []\n        size = len(l) + len(r)\n        for k in range(0, size):\n            lSentinel = i == len(l)\n            rSentinel = j == len(r)\n            if i == len(l):\n                arr.append(r[j])\n                j += 1\n            elif j == len(r):\n                arr.append(l[i])\n                i += 1\n            elif l[i] <= r[j]:\n                arr.append(l[i])\n                i += 1\n            else:\n                arr.append(r[j])\n                j += 1\n\n        return arr\n\n    def merge_sort(self, src):\n        n = len(src) / 2\n        l = src[0:n]\n        r = src[n:]\n        if len(l) > 1:\n            l = self.merge_sort(l)\n        if len(r) > 1:\n            r = self.merge_sort(r)\n\n        src = self.merge(l, r)\n        return src\n\n    def insertion_sort(self, src):\n        j = 1\n        for j in range(1, len(src)):\n            i = j - 1\n            key = src[j]\n        while i >= 0 and src[i] >= key:\n            src[i + 1] = src[i]\n            i = i - 1\n            src[i + 1] = key\n        return src\n```\n\n![](_static/speed_sort_cfg.svg)\n![](docs/source/_static/speed_sort_cfg.svg)\n\nIt can also be used interactively with pudb3.\n\n![](_static/pudb3-py2cfg.gif)\n![](docs/source/_static/pudb3-py2cfg.gif)\n\n### More examples\nAfter cloning, see `./examples/` for some code snippets to run this on.\nTo generate `_cfg.svg` or `_cfg.png` or `_cfg.pdf` images for each example, clone this repo and run the following command in the repo root directory:\n```sh\ngit clone repourl\ncd intorepodirectory\n./generate_examples.sh\n```\n\n## Installation via pip3\nNote: installation is not required, but is handy.\n\nTo install simply run\n```sh\npip3 install py2cfg --user\n```\n\nor clone this repo and pip install locally\n```\ngit clone <url>\ncd intoprojectdirectory\npip3 install . --user\n```\n\n## Usage\nIt can be used three ways:\n\n### Run via shell command\nIf you have installed, then the default command is py2cfg:\n```py\npy2cfg <file.py>\n``` \nThis will create a `<file>_cfg.svg` file, which contains the colored cfg of the file.\nIf you don't want to install via pip, the innards of the py2cfg command can be run right from the repo:\n\n### Run with pudb3\n```bash\npy2cfg <file.py> --debug\n```\n\n### Via wrapper\nIf you have not installed, then you can run a script present in the repo, `py2cfg/_runner.py`, to directly generate a CFG of a Python program and visualise it:\n```sh\ncd intoreporootdir\npython3 py2cfg/_runner.py path_to_my_code.py\n```\n\n### Via import\nWhether or not you have installed (easier if you have), to use py2cfg in your own python code, simply import the module in your Python interpreter or program.\nThen use the `py2cfg.CFGBuilder` class to build CFGs. \nFor example, to build the CFG of a program defined in a file with the path `./example.py`, the following code can be used:\n\n```py\nfrom py2cfg import CFGBuilder\n\ncfg = CFGBuilder().build_from_file('example', './example.py')\n```\n\nThis returns the CFG for the code in `./example.py` in the `cfg` variable. \nThe first parameter of `build_from_file` is the desired name for the CFG, and the second one is the path to the file containing the source code.\nThe produced CFG can then be visualised with:\n\n```py\ncfg.build_visual('exampleCFG', 'pdf')\n```\n\nThe first paramter of `build_visual` is the desired name for the DOT file produced by the method, and the second one is the format to use for the visualisation.\n\n\n# Contributing\n\n## Issues\nModifications and improvements to this project are driven via Gitlab-Issues.\nCheck them out to create one, or fix one.\n\n## Unit tests \nOur minimal tests are run via Gitlab-CI. \nMake sure you don't break them!\n\n## Type hinting\nNote: any new additions to the project should adhere to type-hinting standards enforced by:\n```sh\nmypy --strict --disallow-any-explicit *.py\n```\n* [ ] This is a current issue -- yes we need to fix some things...\n\n## Style\nTo maximize the ability of version control to pin down changes, and keep the style consistent, before you make any commit to the project, run this strict code auto-formatter:\n```sh\nblack py2cfg/*.py\n```\nDo so automatically with pre-commit:\nhttps://black.readthedocs.io/en/stable/version_control_integration.html\n\n# Project history\nNote: py2cfg took it's original inspiration from an older cfg project, though has long since diverged:\n* https://github.com/coetaur0/staticfg\n* https://pypi.org/project/staticfg/\n",
    'author': 'Joe Studer',
    'author_email': 'jmsxw4@mst.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/classroomcode/py2cfg',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# Since the VM's at MST uses a distributed file system, multiple paths may refer
# to the same directory. For example
# /usr/local/.../<SSO>/home
# /mnt/dfs/.../<SSO>/linuxhome
# Both refer to the same directory. For reasons quite literally beyond the scope
# of just py2cfg, we need the /usr/local version
import os
from typing import Callable


def _is_dfs() -> bool:
    # on dfs, os.getcwd() will always return the /mnt/dfs directory
    curdir = os.getcwd()
    homedir = os.environ[
        "HOME"
    ]  # but the environment variable will still refer
    # to /usr/local

    # Even if we explicitly chdir to homedir...
    os.chdir(homedir)
    testdir = os.getcwd()  # Still returns /mnt directory.
    result = not testdir == homedir
    os.chdir(curdir)
    return result


def _normal_path() -> Callable[[str], str]:
    curdir = os.getcwd()
    os.chdir(os.environ["HOME"])
    dfs_home = os.getcwd()
    os.chdir(curdir)

    def normalizer(path):
        return path.replace(dfs_home, os.environ["HOME"])

    return normalizer


# Don't change the path if we're not on MST's VM.
normalizer = _normal_path() if _is_dfs() else lambda path: path
getcwd = lambda: normalizer(os.getcwd())
